function Bone(x, y, angle, dof)
    return {x = x, y = y, angle = angle, dof = dof or 180}
end

function BoneWorld(x, y, angle)
    return {x = x, y = y, angle = angle, cosAngle = math.cos(angle), sinAngle = math.sin(angle)}
end

Chain = {}

function Chain.new(bones)
    local o = {}
    
    setmetatable(o, {__index = Chain})
    
    o.bones = bones
    
    return o
end

function Chain:render()
    love.graphics.push()
    
    for k, v in ipairs(self.bones) do
        love.graphics.rotate(v.angle, 0, 0, 1)
        
        love.graphics.line(0, 0, v.x, 0)
        
        love.graphics.translate(v.x, v.y)
    end
    
    love.graphics.pop()
end

ChainWorld = {}

function ChainWorld.new(chain)
    local o = {}
    
    setmetatable(o, {__index = ChainWorld})
    
    o.bones = {}
    
    o.bones[1] = BoneWorld(chain.bones[1].x, chain.bones[1].y, chain.bones[1].angle)
    
    for i = 2, #chain.bones do
        local prev = o.bones[i - 1]
        local curr = chain.bones[i]
        
        o.bones[i] = BoneWorld(
            prev.x + prev.cosAngle * curr.x - prev.sinAngle * curr.y,
            prev.y + prev.sinAngle * curr.x - prev.cosAngle * curr.y,
            prev.angle + curr.angle
        )
    end
    
    return o
end

function ChainWorld:render()
    for i = #self.bones, 2, -1 do
        love.graphics.line(self.bones[i].x, self.bones[i].y, self.bones[i - 1].x, self.bones[i - 1].y)
    end
end

function simplifyAngle(angle)
    angle = angle % (2 * math.pi)
    
    if  angle < -math.pi then
        angle = angle + (2 * math.pi)
    elseif angle > math.pi then
        angle = angle - 2 * math.pi
    end
    
    return angle
end

function CCD(chain, target, threshold)
    local epsilon = 0.0001
    local trivialArcLength  = 0.00001
    local thresholdSqr = threshold ^ 2 
    local worldBones = ChainWorld.new(chain).bones
    local endX = worldBones[#worldBones].x
    local endY = worldBones[#worldBones].y
    local modified = false
    
    for i = #worldBones - 1, 1, -1 do
        local currToEndX = endX - worldBones[i].x
        local currToEndY = endY - worldBones[i].y
        local currToEndMag = math.sqrt(currToEndX * currToEndX + currToEndY * currToEndY)
        
        local currToTargetX = target.x - worldBones[i].x
        local currToTargetY = target.y - worldBones[i].y
        local currToTargetMag = math.sqrt(currToTargetX * currToTargetX + currToTargetY * currToTargetY)
 
        local cosRotAng
        local sinRotAng
        local endTargetMag = currToEndMag * currToTargetMag
        
        if endTargetMag <= epsilon then
            cosRotAng = 1
            sinRotAng = 0
        else
            cosRotAng = (currToEndX*currToTargetX + currToEndY*currToTargetY) / endTargetMag;
            sinRotAng = (currToEndX*currToTargetY - currToEndY*currToTargetX) / endTargetMag;
        end
        
        local rotAng = math.acos(math.max(-1, math.min(1, cosRotAng)))
        
        if sinRotAng < 0 then
            rotAng = -rotAng
        end

        endX = worldBones[i].x + cosRotAng*currToEndX - sinRotAng*currToEndY
        endY = worldBones[i].y + sinRotAng*currToEndX + cosRotAng*currToEndY
        
        chain.bones[i].angle = simplifyAngle(chain.bones[i].angle + rotAng)
        
        chain.bones[i].angle = math.min(math.max(chain.bones[i].angle, math.rad(-chain.bones[i].dof)), math.rad(chain.bones[i].dof))
 
        local endToTargetX = target.x - endX
        local endToTargetY = target.y - endY
        
        if endToTargetX ^ 2 + endToTargetY ^ 2 <= thresholdSqr then
            return "success"
        end
        
        if not modifiedBones and math.abs(rotAng)*currToEndMag > trivialArcLength then
            modifiedBones = true
        end
    end
    
    if modifiedBones then
        return "processing"
    else
        return "failed"
    end
end

function love.load()
    chain = Chain.new({
        Bone(0, 0, 0),
        Bone(50, 0, 0, 45),
        Bone(50, 0, math.rad(90), 45),
        Bone(50, 0, math.rad(90), 90),
        Bone(50, 0, math.rad(90), 120),
    })
    
    chain2 = Chain.new({
        Bone(0, 0, 0),
        Bone(25, 0, 0),
        Bone(25, 0, math.rad(90)),
        Bone(70, 0, math.rad(90)),
        Bone(25, 0, math.rad(90)),
    })
    
    worldChain = ChainWorld.new(chain)
    worldChain2 = ChainWorld.new(chain2)
end

function love.update(dt)
    CCD(chain, {x = love.mouse.getX() - 400, y = love.mouse.getY() - 300}, 5)
    CCD(chain, {x = love.mouse.getX() - 400, y = love.mouse.getY() - 300}, 5)
    worldChain = ChainWorld.new(chain)
    
    CCD(chain2, {x = love.mouse.getX() - 500, y = love.mouse.getY() - 200}, 5)
    CCD(chain2, {x = love.mouse.getX() - 500, y = love.mouse.getY() - 200}, 5)
    worldChain2 = ChainWorld.new(chain2)
end

function love.draw()
    love.graphics.push()
    love.graphics.translate(400, 300)
    worldChain:render()
    love.graphics.pop()
    
    love.graphics.push()
    love.graphics.translate(500, 200)
    worldChain2:render()
    love.graphics.pop()
end

function love.mousepressed(x, y, button)
    
end

function love.mousereleased(x, y, button)

end


function love.keypressed(key, unicode)
    if unicode == 27 then
        love.event.quit()
    end
end

function love.keyreleased(key, unicode)
    
end

function love.focus(f)
end

function love.quit()
end